#include "action.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "mess.h"
#include "dict.h"
#include "search.h"
#include "utils.h"

static int good(int c)
{
  if ((c >= 'a' && c <= 'z') ||
      (c >= 'A' && c <= 'Z') ||
      (c >= '0' && c <= '9') ||
      c == '+' || c == '-' || c == '/' || c == '*' || c == '_' ||
      c == '"' || c == '\'' || c == '.' || c == ',' || c == ';' ||
      c == ' ' || c == '\n' || c == ':' || c == '\\' ||
      c == '(' || c == ')' || c == '|' || c == '?' || c == '&' ||
      c == '$' || c == '#' || c == '>' || c == '`' || c == '@' ||
      c == '!' || c == '\a' /* bell (7) */ || c == '[' || c == ']' ||
      c == '~' || c == '%' || c == '=' || c == '}' || c == '{' || c == '^' ||
      c == '<' || c == '\t')
    return 1;
  printf("#bad char (%d)\n", c);
  return 0;
}

void check_ascii(char *cables, int skip_euro, unsigned long pos)
{
  FILE *f;
  int c;
  int toget_s[] = { 0x80, 0x99, 0x73, 0 };
  printf("checking asciiness of %s, starting at %lu\n", cables, pos);
  f = fopen(cables, "rb"); if (!f) { perror(cables); exit(1); }
  if (fseek(f, pos, SEEK_SET)) { perror(cables); exit(1); }
  while (1) {
    c = fgetc(f);
    if (c == EOF) break;
    if (!good(c)) {
      if (skip_euro && c == 0xe2) {
        c = fgetc(f);
        if (c != 0x82) { ungetc(c, f); goto next_e2; }
        c = fgetc(f);
        if (c != 0xac) goto bad;
        printf("#skipping euro sign at pos %lu\n", pos);
        printf("./an -patch %lu 1 101 cables.csv\n", pos);
        printf("./an -patch %lu 2 32 cables.csv\n", pos+1);
        pos += 2;
      } else if (skip_euro && c == 0xc2) {
        int i;
        int toget[] = { 0xa0, 0xc2, 0xa0, 0xc2, 0xa0, 0 };
        for (i = 0; toget[i]; i++) if (fgetc(f) != toget[i]) goto bad;
        printf("#skipping C2 A0 C2 A0 C2 A0 at %lu\n", pos);
        printf("./an -patch %lu 6 32 cables.csv\n", pos);
        pos += i;
      } else if (skip_euro && c == 0xe2) {
        int i;
        int c;
next_e2:
        for (i = 0; toget_s[i]; i++) if ((c = fgetc(f)) != toget_s[i]) {
          if (i == 1 && c == 0x9c) goto op_quote;
          if (i == 1 && c == 0x9d) goto cl_quote;
          if (i == 2 && c == 'S') continue; /* 'S is like 's */
          goto bad;
        }
        if (i == 3) {
          printf("#skipping \"'s\" at %lu\n", pos);
          printf("./an -patch %lu 1 39 cables.csv\n", pos);
          printf("./an -patch %lu 1 115 cables.csv\n", pos+1);
          printf("./an -patch %lu 2 32 cables.csv\n", pos+2);
        }
        if (0) {
op_quote: printf("#skipping (open) \" at %lu\n", pos); i++;
          printf("./an -patch %lu 1 32 cables.csv\n", pos);
          printf("./an -patch %lu 1 92 cables.csv\n", pos+1);
          printf("./an -patch %lu 1 34 cables.csv\n", pos+2);
        }
        if (0) {
cl_quote: printf("#skipping (close) \" at %lu\n", pos); i++;
          printf("./an -patch %lu 1 92 cables.csv\n", pos);
          printf("./an -patch %lu 1 34 cables.csv\n", pos+1);
          printf("./an -patch %lu 1 32 cables.csv\n", pos+2);
        }
        pos += i;
      }else {
bad:
        printf("non ascii char found at %lu\n", pos); exit(1);
      }
    }
    pos++;
  }
  fclose(f);
  printf("full ascii file\n");
}

void patch(char *cables, unsigned long pos, unsigned long length, int val)
{
  char c = val;
  FILE *f;
  printf("patching %s at %lu on length %lu, char '%c' (%d)\n",
         cables, pos, length, val, val);
  f = fopen(cables, "r+"); if (!f) { perror(cables); exit(1); }
  if (fseek(f, pos, SEEK_SET)) { perror(cables); exit(1); }
  while (length) {
    if (fwrite(&c, 1, 1, f) != 1) { perror(cables); exit(1); }
    length--;
  }
  if (fclose(f)) { perror(cables); exit(1); }
  printf("patch ok\n");
}

void a_index(char *cables, char *index)
{
  FILE *cf;
  FILE *idf;
  mess m;
  int n = 0;
  long pos;
  uint32_t ppp;
  unsigned char z[4];

  idf = fopen(index, "wb"); if (!idf) { perror(index); exit(1); }
  cf = fopen(cables, "rb"); if (!cf) { perror(cables); exit(1); }
  while (1) {
    pos = ftell(cf);
    if (!load_mess(cf, &m)) break;
    close_mess(&m);
printf("%d\r", n);
    ppp = pos;
    z[0] = ppp & 255;
    z[1] = (ppp>>8) & 255;
    z[2] = (ppp>>16) & 255;
    z[3] = (ppp>>24) & 255;
    if (fwrite(z, 4, 1, idf) != 1) { perror(index); exit(1); }
    n++;
  }
  if (fclose(idf)) { perror(index); exit(1); }

  printf("%d cables\n", n);
}

void get(char *cables, char *index, int m, int fields)
{
  FILE *idf;
  FILE *cf;
  unsigned char z[4];
  int pos;
  mess mm;

  if (fields == -1) V("get message %d\n", m);

  idf = fopen(index, "r"); if (!idf) { perror(index); exit(1); }
  if (fseek(idf, (m-1)*4, SEEK_SET)) { perror(index); exit(1); }
  if (fread(z, 4, 1, idf) != 1) { perror(index); exit(1); }
  fclose(idf);
  pos = z[0] | (z[1] << 8) | (z[2] << 16) | (z[3] << 24);
  if (fields == -1) V("message is at offset %d\n", pos);
  cf = fopen(cables, "rb"); if (!cf) { perror(cables); exit(1); }
  if (fseek(cf, pos, SEEK_SET)) { perror(cables); exit(1); }
  if (!load_mess(cf, &mm)) { printf("error getting mess %d\n", m); exit(1); }
  if (fields == -1) {
    printf("F1:\n--------\n%s\n--------\n", mm.f1.data);
    printf("F2:\n--------\n%s\n--------\n", mm.f2.data);
    printf("F3:\n--------\n%s\n--------\n", mm.f3.data);
    printf("F4:\n--------\n%s\n--------\n", mm.f4.data);
    printf("F5:\n--------\n%s\n--------\n", mm.f5.data);
    printf("F6:\n--------\n%s\n--------\n", mm.f6.data);
    printf("F7:\n--------\n%s\n--------\n", mm.f7.data);
    printf("F8:\n--------\n%s\n--------\n", mm.f8.data);
  } else {
    if (fields == 1) printf("%s", mm.f1.data);
    if (fields == 2) printf("%s", mm.f2.data);
    if (fields == 3) printf("%s", mm.f3.data);
    if (fields == 4) printf("%s", mm.f4.data);
    if (fields == 5) printf("%s", mm.f5.data);
    if (fields == 6) printf("%s", mm.f6.data);
    if (fields == 7) printf("%s", mm.f7.data);
    if (fields == 8) printf("%s", mm.f8.data);
    printf("\n");
  }
  close_mess(&mm);
  fclose(cf);
}

void words(char *cables, char *index, int fields)
{
  FILE *f;
  int *id;
  int i;
  mess m;
  dict d;

  id = load_index(index);
  if (!id) exit(1);

  f = fopen(cables, "rb"); if (!f) { perror(cables); exit(1); }

  init_dict(&d);

  for (i = 0; i < 251287; i++) {
//printf("%d\r", i);
    if (fseek(f, id[i], SEEK_SET))
      { printf("i is %d (id: %d)\n", i, id[i]); perror(cables); exit(1); }
    if (!load_mess(f, &m)) { printf("error with mess %d\n", i+1); exit(1); }
    if (fields & 0x01) dict_populate(&d, (unsigned char *)m.f1.data);
    if (fields & 0x02) dict_populate(&d, (unsigned char *)m.f2.data);
    if (fields & 0x04) dict_populate(&d, (unsigned char *)m.f3.data);
    if (fields & 0x08) dict_populate(&d, (unsigned char *)m.f4.data);
    if (fields & 0x10) dict_populate(&d, (unsigned char *)m.f5.data);
    if (fields & 0x20) dict_populate(&d, (unsigned char *)m.f6.data);
    if (fields & 0x40) dict_populate(&d, (unsigned char *)m.f7.data);
    if (fields & 0x80) dict_populate(&d, (unsigned char *)m.f8.data);
    close_mess(&m);
  }
//printf("\n");

  printf("%d words, %d unique words, %d collisions\n",
         d.total_words, d.count, d.h.collision);
  printf("COUNT WORD:\n");
  for (i = 0; i < d.count; i++) {
    printf("%d %s\n",
           d.words[i].count,
           d.h.keys + d.words[i].key);
  }
  close_dict(&d);
}

/* ugly global, but that's qsort... */
static dict *dico;
static int cmp(const void *a, const void *b)
{
  const int *aa = a;
  const int *bb = b;
  int va = *aa;
  int vb = *bb;
  const char *sa = dico->h.keys + dico->words[va].key;
  const char *sb = dico->h.keys + dico->words[vb].key;
  return strcmp(sa, sb);
}

void do_dict(char *cables, char *index, char *dict_file, char *table_file,
             int dict_fields)
{
  FILE *f;
  int *id;
  int i;
  int j;
  mess m;
  dict d;
  int *order;
  FILE *f_dict;
  FILE *f_table;
  uint32_t offset;

  id = load_index(index);
  if (!id) exit(1);

  f = fopen(cables, "rb"); if (!f) { perror(cables); exit(1); }

  init_dict(&d);

  for (i = 0; i < 251287; i++) {
    if (fseek(f, id[i], SEEK_SET))
      { printf("i is %d (id: %d)\n", i, id[i]); perror(cables); exit(1); }
    if (!load_mess(f, &m)) { printf("error with mess %d\n", i+1); exit(1); }
    if (dict_fields&0x01) dict_word_to_doc(&d, (unsigned char *)m.f1.data, i);
    if (dict_fields&0x02) dict_word_to_doc(&d, (unsigned char *)m.f2.data, i);
    if (dict_fields&0x04) dict_word_to_doc(&d, (unsigned char *)m.f3.data, i);
    if (dict_fields&0x08) dict_word_to_doc(&d, (unsigned char *)m.f4.data, i);
    if (dict_fields&0x10) dict_word_to_doc(&d, (unsigned char *)m.f5.data, i);
    if (dict_fields&0x20) dict_word_to_doc(&d, (unsigned char *)m.f6.data, i);
    if (dict_fields&0x40) dict_word_to_doc(&d, (unsigned char *)m.f7.data, i);
    if (dict_fields&0x80) dict_word_to_doc(&d, (unsigned char *)m.f8.data, i);
    close_mess(&m);
  }

  /* order words */
  dico = &d;
  order = malloc(sizeof(int) * d.count);
  if (!order) { perror("malloc"); exit(1); }
  for (i = 0; i < d.count; i++) order[i] = i;
  qsort(order, d.count, sizeof(int), cmp);

  printf("%d words, %d unique words, %d collisions\n",
         d.total_words, d.count, d.h.collision);
#if 0
  printf("COUNT WORD: DOCUMENT ... DOCUMENT\n");
  for (i = 0; i < d.count; i++) {
    int ii = order[i];
    printf("%d %s:",
           d.words[ii].count,
           d.h.keys + d.words[ii].key);
    for (j = 0; j < d.words[ii].size; j++)
      printf(" %d", d.words[ii].indoc[j]+1);
    printf("\n");
  }
#endif

  /* create dict & table */
  f_dict = fopen(dict_file, "wb");
  if (f_dict == NULL) { perror(dict_file); exit(1); }
  fprintf(f_dict, "%d\n", d.count);
  for (i = 0; i < d.count; i++) {
    int ii = order[i];
    if (fprintf(f_dict, "%s\n", d.h.keys + d.words[ii].key) == -1)
      { perror(dict_file); exit(1); }
  }
  if (fclose(f_dict)) { perror(dict_file); exit(1); }

  f_table = fopen(table_file, "wb");
  if (f_table == NULL) { perror(table_file); exit(1); }
  /* write offset table */
  offset = d.count * sizeof(uint32_t);
  for (i = 0; i < d.count; i++) {
    int ii = order[i];
    char c[4];
    c[0] = offset & 255;
    c[1] = (offset >> 8) & 255;
    c[2] = (offset >> 16) & 255;
    c[3] = (offset >> 24) & 255;
    if (fwrite(c, 4, 1, f_table) != 1) { perror(table_file); exit(1); }
    offset += d.words[ii].size * 4;
  }
  /* write docs */
  for (i = 0; i < d.count; i++) {
    char c[4];
    uint32_t doc;
    int ii = order[i];
    for (j = 0; j < d.words[ii].size; j++) {
      doc = d.words[ii].indoc[j];
    c[0] = doc & 255;
    c[1] = (doc >> 8) & 255;
    c[2] = (doc >> 16) & 255;
    c[3] = (doc >> 24) & 255;
    if (fwrite(c, 4, 1, f_table) != 1) { perror(table_file); exit(1); }
    }
  }
  if (fclose(f_table)) { perror(table_file); exit(1); }

  close_dict(&d);
}

static char map[251287];
void word_search(char *cables, char *index, char *dict_file, char *table_file,
                 char *word)
{
  word_list w;
  int n;
  int i;
  int count = 0;
  word = bigcap(word);
  load_words(dict_file, table_file, &w);
  n = search_word(&w, word);
  if (n == -1) { printf("%s: word not found\n", word); exit(1); }
  memset(map, 0, 251287);
  feed_map(map, &w, n, table_file);
  for (i = 0; i < 251287; i++) if (map[i]) count++;
  V("%s\n%d\n", word, count);
  for (i = 0; i < 251287; i++) if (map[i]) printf(" %d", i+1);
  printf("\n");
  free(word);
}

void search(char *cables, char *index, char *dict_file, char *table_file,
            char *expression)
{
  char *r;
  int i;
  int count = 0;
  expression = bigcap(expression);
  r = complex_search(cables, index, dict_file, table_file, expression);
  for (i = 0; i < 251287; i++) if (r[i]) count++;
  V("%s\n%d\n", expression, count);
  for (i = 0; i < 251287; i++) if (r[i]) printf(" %d", i+1);
  printf("\n");
  free(r);
  free(expression);
}
