#include "dict.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "utils.h"

void init_dict(dict *d)
{
  int maxentries;
  maxentries = 16 * 1024 * 1024;
  d->words = malloc(maxentries * sizeof(word));
  if (d->words == NULL) { perror("malloc"); exit(1); }
  memset(d->words, 0, sizeof(word) * maxentries);
  d->count = 0;
  d->total_words = 0;
  init_buffer(&d->curword, 512, 128);
  init_hash(&d->h, maxentries);
  spacify(d->isspace);
  upperify(d->upper);
}

void close_dict(dict *d)
{
  close_buffer(&d->curword);
  close_hash(&d->h);
  /* we should free words' data, but who cares... */
}

static const unsigned char *skip_space(dict *d, const unsigned char *m)
{
  while (*m && d->isspace[*m]) m++;
  return m;
}

void dict_populate(dict *d, const unsigned char *m)
{
  record v;
  while (*m) {
    m = skip_space(d, m);
    d->curword.size = 0;
    while (!d->isspace[*m]) {
      bchar(&d->curword, d->upper[*m]);
      m++;
    }
    if (d->curword.size == 0) continue;
    bchar(&d->curword, 0);
    d->total_words++;
    v = hash_get(&d->h, d->curword.data);
    if (d->words[v.val].count == 0) d->count++;
    d->words[v.val].count++;
    d->words[v.val].key = v.key;
  }
}

void dict_word_to_doc(dict *d, const unsigned char *m, int doc)
{
  record v;
  word *w;
  while (*m) {
    m = skip_space(d, m);
    d->curword.size = 0;
    while (!d->isspace[*m]) {
      bchar(&d->curword, d->upper[*m]);
      m++;
    }
    if (d->curword.size == 0) continue;
    bchar(&d->curword, 0);
    d->total_words++;
    v = hash_get(&d->h, d->curword.data);
    if (d->words[v.val].count == 0) d->count++;
    d->words[v.val].count++;
    d->words[v.val].key = v.key;
    /* the word is in document 'doc', store that info if not there yet */
    w = &d->words[v.val];
    if (w->size != 0 && w->indoc[w->size-1] == doc) continue;
    if (w->maxsize == w->size) {
      if (w->maxsize == 0) w->maxsize = 16; else w->maxsize += 256;
      if (w->maxsize == 16)
        w->indoc = malloc(sizeof(int) * w->maxsize);
      else
        w->indoc = realloc(w->indoc, sizeof(int) * w->maxsize);
      if (w->indoc == NULL) { perror("malloc"); exit(1); }
    }
    w->indoc[w->size] = doc;
    w->size++;
  }
}
