#include "hash.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

void init_hash(hash *h, int size)
{
  int i;

  h->data = malloc(size * sizeof(record));
  if (h->data == NULL) { perror("malloc"); exit(1); }
  memset(h->data, 0, size * sizeof(record));
  h->size = 0;
  h->maxsize = size;

  h->keys = malloc(32 * 1024 * 1024);
  if (h->keys == NULL) { perror("malloc"); exit(1); }
  h->keysize = 1;
  h->keymaxsize = 32 * 1024 * 1024;
  h->collision = 0;

  for (i = 0; i < 256*256; i++) {
    h->h[i/256][i%256] = rand();
  }
}

void close_hash(hash *h)
{
  free(h->data);
  free(h->keys);
}

static int hash_calc(hash *h, const unsigned char *k)
{
#if 0
  extern uint32_t jhash(unsigned char *, uint32_t, uint32_t);
  return jhash((unsigned char *)k, strlen((const char *)k), 0) & (h->maxsize-1);
#else
  /* (very slightly) faster than Jenkins and (very slightly too) less
   * collisions for ten minutes of thinking, go figure...
   */
  int pos = 0;
  int v = 0;
  while (*k) {
    v += h->h[*k][pos];
    pos++;
    pos &= 255;
    k++;
  }
  v &= h->maxsize-1;
  return v;
#endif
}

record hash_get(hash *h, const char *k)
{
  int k_hash = hash_calc(h, (const unsigned char *)k);
  int l;
  int col = 0;
  while (h->data[k_hash].key) {
    if (!strcmp(k, h->keys + h->data[k_hash].key)) return h->data[k_hash];
    col++;
    k_hash++;
  }
  h->collision += col;
  /* not there, put it */
  l = strlen(k);
  if (h->size == h->maxsize) { printf("hash full\n"); exit(1); }
  if (h->keysize + l + 1 > h->keymaxsize)
    { printf("hash keys full (implement resize, lazy guy...)\n"); exit(1); }
  h->data[k_hash].val = h->size;
  h->data[k_hash].key = h->keysize;
  memcpy(h->keys + h->keysize, k, l+1);
  h->size++;
  h->keysize += l+1;
  return h->data[k_hash];
}
