#include <string.h>
#include <stdio.h>
#include <stdlib.h>

int verbose;

#include "action.h"

static char *prog;
void usage(void)
{
  printf("usage: %s <action> [options] <cable file>\n", prog);
  printf("actions:\n");
#if 0
  /* noone will do that, no need to advertize it */
  printf("    -check-ascii : check that the file is ASCII only\n");
  printf("    -patch <offset> <length> <value> : patch the file\n");
#endif
  printf("    -build-index : create the index file\n");
  printf("    -get-cable <cable number> : get the wanted cable\n");
  printf("    -words : get the words, dump stats\n");
  printf("    -dict : create a dictionary (dict file & table file)\n");
  printf("    -word-search <word> : search a word, return docs # it's in\n");
  printf("    -search <expression> : complex search, return docs #\n");
  printf("options:\n");
  printf("    -index-file <file> : index file is <index>\n");
  printf("    -dict-file <file> : dict file is <index>\n");
  printf("    -table-file <file> : table file is <index>\n");
  printf("    -get-cable-field <field> : print only field <field> (in [1-8])\n");
#if 0
  printf("    -check-skip-euro : don't stop on euro sign (and more)\n");
  printf("    -check-start-at <offset> : start check at given offset\n");
#endif
  printf("    -dict-fields XXXXXXXX: what fields to use for the dictionary\n");
  printf("        example: for fields 3&8 pass: 10000100 (default: all, 11111111)\n");
  printf("    -verbose : be verbose\n");
  printf("http://sed.free.fr/cables\n");
  printf("public domain software\n");
  exit(0);
}

enum action { NOTHING, CHECK_ASCII, PATCH, INDEX, GET_MESSAGE, WORDS, DICT,
              WORD_SEARCH, SEARCH };

int main(int n, char **v)
{
  int i;
  enum action action = NOTHING;
  char *cables = NULL;
  unsigned long patch_offset;
  unsigned long patch_length;
  int patch_value;
  int check_skip_euro = 0;
  unsigned long check_offset = 0;
  char *index_file = NULL;
  int message_number = -1;
  int get_message_fields = -1;
  char *dict_file = NULL;
  char *table_file = NULL;
  char *searched_word = NULL;
  char *searched_expression = NULL;
  int dict_fields = 0xff;

  prog = v[0];

  for (i = 1; i < n; i++) {
    /* actions */
    if (!strcmp(v[i], "-check-ascii")) {
      if (action != NOTHING) usage();
      action = CHECK_ASCII;
      continue;
    }
    if (!strcmp(v[i], "-patch")) {
      if (action != NOTHING) usage();
      if (i >= n-3) usage();
      if (sscanf(v[i+1], "%lu", &patch_offset) != 1) usage();
      if (sscanf(v[i+2], "%lu", &patch_length) != 1) usage();
      if (sscanf(v[i+3], "%d", &patch_value) != 1) usage();
      i+=3;
      action = PATCH;
      continue;
    }
    if (!strcmp(v[i], "-build-index")) {
      if (action != NOTHING) usage();
      action = INDEX;
      continue;
    }
    if (!strcmp(v[i], "-get-cable")) {
      if (action != NOTHING) usage();
      if (i >= n-1) usage();
      if (sscanf(v[i+1], "%d", &message_number) != 1) usage();
      i++;
      action = GET_MESSAGE;
      continue;
    }
    if (!strcmp(v[i], "-words")) {
      if (action != NOTHING) usage();
      action = WORDS;
      continue;
    }
    if (!strcmp(v[i], "-dict")) {
      if (action != NOTHING) usage();
      action = DICT;
      continue;
    }
    if (!strcmp(v[i], "-word-search")) {
      if (action != NOTHING) usage();
      action = WORD_SEARCH;
      if (i >= n-1) usage();
      searched_word = v[i+1];
      i++;
      continue;
    }
    if (!strcmp(v[i], "-search")) {
      if (action != NOTHING) usage();
      action = SEARCH;
      if (i >= n-1) usage();
      searched_expression = v[i+1];
      i++;
      continue;
    }
    /* options */
    if (!strcmp(v[i], "-index-file")) {
      if (index_file != NULL) usage();
      if (i >= n-1) usage();
      index_file = v[i+1];
      i++;
      continue;
    }
    if (!strcmp(v[i], "-dict-file")) {
      if (dict_file != NULL) usage();
      if (i >= n-1) usage();
      dict_file = v[i+1];
      i++;
      continue;
    }
    if (!strcmp(v[i], "-table-file")) {
      if (table_file != NULL) usage();
      if (i >= n-1) usage();
      table_file = v[i+1];
      i++;
      continue;
    }
    if (!strcmp(v[i], "-get-cable-field")) {
      if (i >= n-1) usage();
      if (sscanf(v[i+1], "%d", &get_message_fields) != 1) usage();
      if (get_message_fields != 1 && get_message_fields != 2 &&
          get_message_fields != 3 && get_message_fields != 4 &&
          get_message_fields != 5 && get_message_fields != 6 &&
          get_message_fields != 7 && get_message_fields != 8) usage();
      i++;
      continue;
    }
    if (!strcmp(v[i], "-check-skip-euro")) {
      check_skip_euro = 1;
      continue;
    }
    if (!strcmp(v[i], "-check-start-at")) {
      if (i >= n-1) usage();
      if (sscanf(v[i+1], "%lu", &check_offset) != 1) usage();
      i++;
      continue;
    }
    if (!strcmp(v[i], "-dict-fields")) {
      int f = 0x80, j;
      if (i >= n-1) usage();
      if (strlen(v[i+1]) != 8) usage();
      dict_fields = 0;
      for (j = 0; j < 8; j++) {
        if (v[i+1][j] == '1') dict_fields |= f;
        else if (v[i+1][j] != '0') usage();
        f >>= 1;
      }
      i++;
      continue;
    }
    if (!strcmp(v[i], "-verbose")) {
      verbose = 1;
      continue;
    }
    if (cables) usage();
    cables = v[i];
  }
  if (cables == NULL) usage();

  switch (action) {
  case NOTHING: break;
  case CHECK_ASCII: check_ascii(cables, check_skip_euro, check_offset); break;
  case PATCH: patch(cables, patch_offset, patch_length, patch_value); break;
  case INDEX: a_index(cables, index_file); break;
  case GET_MESSAGE: if (index_file == NULL) usage();
    get(cables, index_file, message_number, get_message_fields); break;
  case WORDS: if (index_file == NULL) usage();
    words(cables, index_file, dict_fields); break;
  case DICT:
    if (index_file == NULL || dict_file == NULL || table_file == NULL) usage();
    do_dict(cables, index_file, dict_file, table_file, dict_fields); break;
  case WORD_SEARCH:
    if (index_file == NULL || dict_file == NULL || table_file == NULL) usage();
    word_search(cables, index_file, dict_file, table_file, searched_word);
    break;
  case SEARCH:
    if (index_file == NULL || dict_file == NULL || table_file == NULL) usage();
    search(cables, index_file, dict_file, table_file, searched_expression);
    break;
  }

  return 0;
}
