#include "mess.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

void blanks(FILE *f)
{
  int c;
  while (isspace((c = fgetc(f))));
  ungetc(c, f);
}

#define g(f) fgetc(f)

int gs(FILE *f, buffer *b)
{
  int c;
  while (1) {
    c = g(f);
    if (c == EOF) return 0;
    if (c == '"') { ungetc(c, f); return 1; }
    if (c == '\\') {
      c = g(f);
      if (c != '"' && c != '\'' && c != '\\')
        { printf("bad escape char %c (%d)\n", c, c); exit(1); }
    }
    bchar(b, c);
  }
}

int load_mess(FILE *f, mess *m)
{
  init_buffer(&m->f1, 16, 16);
  init_buffer(&m->f2, 16, 16);
  init_buffer(&m->f3, 16, 16);
  init_buffer(&m->f4, 16, 16);
  init_buffer(&m->f5, 16, 16);
  init_buffer(&m->f6, 16, 16);
  init_buffer(&m->f7, 512, 128);
  init_buffer(&m->f8, 2048, 1024);
  if (g(f) != '"') return 0;
  if (!gs(f, &m->f1) || g(f) != '"' || g(f) != ',' || g(f) != '"') return 0;
  if (!gs(f, &m->f2) || g(f) != '"' || g(f) != ',' || g(f) != '"') return 0;
  if (!gs(f, &m->f3) || g(f) != '"' || g(f) != ',' || g(f) != '"') return 0;
  if (!gs(f, &m->f4) || g(f) != '"' || g(f) != ',' || g(f) != '"') return 0;
  if (!gs(f, &m->f5) || g(f) != '"' || g(f) != ',' || g(f) != '"') return 0;
  if (!gs(f, &m->f6) || g(f) != '"' || g(f) != ',' || g(f) != '"') return 0;
  if (!gs(f, &m->f7) || g(f) != '"' || g(f) != ',' || g(f) != '"') return 0;
  if (!gs(f, &m->f8) || g(f) != '"') return 0;
  blanks(f);
  bchar(&m->f1, 0);
  bchar(&m->f2, 0);
  bchar(&m->f3, 0);
  bchar(&m->f4, 0);
  bchar(&m->f5, 0);
  bchar(&m->f6, 0);
  bchar(&m->f7, 0);
  bchar(&m->f8, 0);
  return 1;
}

void close_mess(mess *m)
{
  close_buffer(&m->f1);
  close_buffer(&m->f2);
  close_buffer(&m->f3);
  close_buffer(&m->f4);
  close_buffer(&m->f5);
  close_buffer(&m->f6);
  close_buffer(&m->f7);
  close_buffer(&m->f8);
}

int *load_index(char *index)
{
  unsigned char c[4];
  int i;
  FILE *idx;
  int *ret = malloc(251287 * sizeof(int));
  if (!ret) { perror("malloc"); exit(1); }
  idx = fopen(index, "rb"); if (idx == NULL) { perror(index); exit(1); }
  for (i = 0; i < 251287; i++) {
    if (fread(c, 4, 1, idx) != 1) { perror(index); exit(1); }
    ret[i] = (c[0]) | (c[1] << 8) | (c[2] << 16) | (c[3] << 24);
  }
  return ret;
}

static const unsigned char *skip_space(const unsigned char *m,
                                       unsigned char *space)
{
  while (*m && space[*m]) m++;
  return m;
}

static void populate(buffer *o, const char *_m, unsigned char *space,
                     unsigned char *upper)
{
  const unsigned char *m = (const unsigned char *)_m;
  m = skip_space(m, space);
  while (*m) {
    while (!space[*m]) { bchar(o, upper[*m]); m++; }
    m = skip_space(m, space);
    if (*m) bchar(o, ' ');
  }
}

void get_raw(char *cables, char *index, int m, buffer *o,
             unsigned char *space, unsigned char *upper)
{
  FILE *idf;
  FILE *cf;
  unsigned char z[4];
  int pos;
  mess mm;

  idf = fopen(index, "r"); if (!idf) { perror(index); exit(1); }
  if (fseek(idf, m*4, SEEK_SET)) { perror(index); exit(1); }
  if (fread(z, 4, 1, idf) != 1) { perror(index); exit(1); }
  fclose(idf);
  pos = z[0] | (z[1] << 8) | (z[2] << 16) | (z[3] << 24);

  cf = fopen(cables, "rb"); if (!cf) { perror(cables); exit(1); }
  if (fseek(cf, pos, SEEK_SET)) { perror(cables); exit(1); }
  if (!load_mess(cf, &mm)) { printf("error getting mess %d\n", m); exit(1); }

  o->size = 0;

  bchar(o, ' ');
  populate(o, mm.f1.data, space, upper);
  bchar(o, ' ');
  populate(o, mm.f2.data, space, upper);
  bchar(o, ' ');
  populate(o, mm.f3.data, space, upper);
  bchar(o, ' ');
  populate(o, mm.f4.data, space, upper);
  bchar(o, ' ');
  populate(o, mm.f5.data, space, upper);
  bchar(o, ' ');
  populate(o, mm.f6.data, space, upper);
  bchar(o, ' ');
  populate(o, mm.f7.data, space, upper);
  bchar(o, ' ');
  populate(o, mm.f8.data, space, upper);
  bchar(o, ' ');
  bchar(o, 0);

  close_mess(&mm);
  fclose(cf);
}
